import asyncio
import logging
from asyncio import iscoroutinefunction
from collections import namedtuple
from typing import Callable, Tuple, Type

from croniter import croniter

from aiomisc import Service
from aiomisc.cron import CronCallback

log = logging.getLogger(__name__)

StoreItem = namedtuple(
    "StoreItem", "callback, spec, shield, suppress_exceptions"
)


class CronService(Service):
    def __init__(self, **kwargs):
        super(CronService, self).__init__(**kwargs)
        self._callbacks_storage = set()

    def register(
            self,
            function: Callable,
            spec: str,
            shield: bool = False,
            suppress_exceptions: Tuple[Type[Exception]] = ()
    ):
        if not iscoroutinefunction(function):
            raise TypeError("function should be a coroutine %r" % function)
        if not croniter.is_valid(spec):
            raise TypeError("Not valid cron spec %r" % spec)

        self._callbacks_storage.add(
            StoreItem(CronCallback(function), spec, shield, suppress_exceptions)
        )

    async def start(self):
        for item in self._callbacks_storage:
            item.callback.start(
                spec=item.spec,
                loop=self.loop,
                shield=item.shield,
                suppress_exceptions=item.suppress_exceptions
            )
        log.info("Cron service %s started", self)

    async def stop(self, exception: Exception = None):
        async def _shutdown(item: StoreItem):
            if item.callback.task:
                await item.callback.task
            item.callback.stop()

        await asyncio.gather(
            *[_shutdown(store) for store in self._callbacks_storage],
            return_exceptions=True,
        )
        log.info("Cron service %s is stopped", self)

    def __str__(self):
        storage = ", ".join(
            "{}: {}".format(item.callback, item.spec) for item in
            self._callbacks_storage
        )
        return "{}({})".format(
            self.__class__.__name__,
            storage,
        )



import logging

import pytest


@pytest.fixture
async def yield_fixture(loop):
    logging.info("Setup")

    f = loop.create_future()
    loop.call_later(0, f.set_result, True)
    await f

    try:
        yield True
    finally:
        f = loop.create_future()
        loop.call_later(0, f.set_result, True)
        await f

        logging.info("Teardown")



import asyncio
import logging
import logging.handlers
import time
from contextlib import suppress
from typing import Union

from prettylog import (
    JSONLogFormatter, LogFormat, color_formatter, create_logging_handler,
    json_formatter,
)
from prettylog import wrap_logging_handler as _wrap_logging_handler

from .thread_pool import run_in_new_thread


def _thread_flusher(
    handler: logging.handlers.MemoryHandler,
    flush_interval: Union[float, int],
    loop: asyncio.AbstractEventLoop,
):
    def has_no_target():
        return True

    def has_target():
        return bool(handler.target)

    is_target = has_no_target

    if isinstance(handler, logging.handlers.MemoryHandler):
        is_target = has_target

    while not loop.is_closed() and is_target():
        with suppress(Exception):
            if handler.buffer:
                handler.flush()

        time.sleep(flush_interval)


def wrap_logging_handler(
    handler: logging.Handler,
    loop: asyncio.AbstractEventLoop = None,
    buffer_size: int = 1024,
    flush_interval: float = 0.1,
):
    loop = loop or asyncio.get_event_loop()

    buffered_handler = _wrap_logging_handler(
        handler=handler, buffer_size=buffer_size,
    )

    run_in_new_thread(
        _thread_flusher, args=(
            buffered_handler, flush_interval, loop,
        ), no_return=True,
    )

    return buffered_handler


def basic_config(
    level: int = logging.INFO,
    log_format: Union[str, LogFormat] = LogFormat.color,
    buffered: bool = True, buffer_size: int = 1024,
    flush_interval: float = 0.2, loop=None, **kwargs
):

    if isinstance(level, str):
        level = getattr(logging, level.upper())

    logging.basicConfig()
    logger = logging.getLogger()
    logger.handlers.clear()

    if isinstance(log_format, str):
        log_format = LogFormat[log_format]

    handler = create_logging_handler(log_format, **kwargs)

    if buffered:
        handler = wrap_logging_handler(
            handler,
            buffer_size=buffer_size,
            flush_interval=flush_interval,
            loop=loop,
        )

    logging.basicConfig(
        level=level,
        handlers=[handler],
    )


__all__ = (
    "basic_config",
    "color_formatter",
    "json_formatter",
    "JSONLogFormatter",
    "LogFormat",
    "wrap_logging_handler",
)



import logging
import re
from types import MappingProxyType
from typing import List  # NOQA

from aiocarbon.protocol.pickle import PickleClient
from aiocarbon.protocol.tcp import TCPClient
from aiocarbon.protocol.udp import UDPClient
from aiocarbon.setup import set_client
from aiocarbon.storage import TotalStorage
from aiomisc.periodic import PeriodicCallback
from aiomisc.service import Service


log = logging.getLogger(__name__)


def strip_carbon_ns(string):
    return re.sub(r"[^\w\d\-]+", "_", string).strip("_").lower()


PROTOCOLS = MappingProxyType({
    "udp": UDPClient,
    "tcp": TCPClient,
    "pickle": PickleClient,
})


class CarbonSender(Service):
    host = "127.0.0.1"          # type: str
    port = 2003                 # type: int
    send_interval = 5           # type: int
    protocol = "udp"            # type: str
    namespace = ""              # type: List[str]
    storage = TotalStorage
    _handle = None              # type: PeriodicCallback

    async def start(self):
        namespace = ".".join(
            strip_carbon_ns(item) for item in self.namespace
        )

        client = PROTOCOLS[self.protocol](
            self.host,
            self.port,
            namespace=namespace,
            storage=self.storage(),
        )

        set_client(client)

        self._handle = PeriodicCallback(client.send)
        self._handle.start(self.send_interval, loop=self.loop)
        log.info(
            "Periodic carbon metrics sender started. Send to %s://%s:%d with "
            "interval %rs", self.protocol, self.host, self.port,
            self.send_interval,
        )

    async def stop(self, *_):
        self._handle.stop()



import logging

from aiomisc import PeriodicCallback, Service


log = logging.getLogger(__name__)


class PeriodicService(Service):

    __required__ = ("interval",)

    interval = None  # type: float # in seconds
    delay = 0  # type: float # in seconds

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.periodic = PeriodicCallback(self.callback)

    async def start(self):
        self.periodic.start(self.interval, delay=self.delay, loop=self.loop)
        log.info("Periodic service %s started", self)

    async def stop(self, err):
        if self.periodic.task:
            await self.periodic.task
        self.periodic.stop()
        log.info("Periodic service %s is stopped", self)

    async def callback(self):
        raise NotImplementedError

    def __str__(self):
        return "{}(interval={},delay={})".format(
            self.__class__.__name__,
            self.interval,
            self.delay,
        )



from .base import Service, ServiceMeta, SimpleServer
from .profiler import Profiler
from .tcp import TCPServer
from .tls import TLSServer
from .tracer import MemoryTracer
from .udp import UDPServer


__all__ = (
    "MemoryTracer",
    "Profiler",
    "Service",
    "ServiceMeta",
    "SimpleServer",
    "TCPServer",
    "TLSServer",
    "UDPServer",
)



import asyncio
import logging
from datetime import datetime
from functools import partial
from typing import Tuple, Type

from croniter import croniter

from . import utils

log = logging.getLogger(__name__)


class CronCallback:
    """
    .. note::

        When the cron function executes longer then execution interval a
        next call will be skipping and warning will be logged.

    """

    __slots__ = (
        "_cb", "_closed", "_task", "_loop", "_handle", "__name",
        "_croniter"
    )

    def __init__(self, coroutine_func, *args, **kwargs):
        self.__name = repr(coroutine_func)
        self._cb = partial(utils.awaitable(coroutine_func), *args, **kwargs)
        self._closed = False
        self._loop = None
        self._handle = None
        self._task = None

        self._croniter = None

    async def _run(self, suppress_exceptions=()):
        try:
            await self._cb()
        except suppress_exceptions:
            return
        except asyncio.CancelledError:
            raise
        except Exception:
            log.exception("Cron task error:")

    def get_next(self):
        if not self._loop or not self._croniter:
            raise asyncio.InvalidStateError
        loop_time = self._loop.time()
        timestamp = datetime.utcnow().timestamp()
        return (
                loop_time +
                (self._croniter.get_next(float) - timestamp)
        )

    def start(
            self,
            spec: str,
            loop=None, *,
            shield: bool = False,
            suppress_exceptions: Tuple[Type[Exception]] = ()
    ):
        if self._task and not self._task.done():
            raise asyncio.InvalidStateError

        self._loop = loop or asyncio.get_event_loop()

        self._croniter = croniter(
            spec, start_time=datetime.utcnow().timestamp()
        )

        self._closed = False

        def cron():
            if self._loop.is_closed():
                return

            if self._task and not self._task.done():
                log.warning("Task %r still running skipping", self)
                call_next()
                return

            del self._task
            self._task = None

            if self._closed:
                return

            runner = utils.shield(self._run) if shield else self._run
            self._task = self._loop.create_task(runner(suppress_exceptions))

            call_next()

        def call_next(*_):
            if self._handle is not None:
                self._handle.cancel()
                del self._handle

            self._handle = self._loop.call_at(self.get_next(), cron)

        self._loop.call_at(
            self.get_next(), self._loop.call_soon_threadsafe, cron
        )

    def stop(self):
        self._closed = True

        if self._task is None:
            self._task = self._loop.create_future()
            self._task.set_exception(RuntimeError("Callback not started"))
        elif not self._task.done():
            self._task.cancel()

        if self._handle:
            self._handle.cancel()

        return self._task

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.__name)

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, self._cb.func.__name__)

    @property
    def task(self):
        return self._task



import asyncio
import socket

import pytest

from aiomisc import Service


class _TestService(Service):
    loop_on_init = None  # type: asyncio.AbstractEventLoop

    async def start(self):
        assert self.loop is asyncio.get_event_loop()


@pytest.fixture()
async def async_sleep(loop):
    f = loop.create_future()
    loop.call_soon(f.set_result, True)
    return await f


@pytest.fixture()
def service(loop: asyncio.AbstractEventLoop):
    return _TestService(loop_on_init=loop)


@pytest.fixture()
def services(service: _TestService, async_sleep):
    return [service]


async def test_loop_fixture(
    service: _TestService,
    loop: asyncio.AbstractEventLoop
):
    assert service.loop is service.loop_on_init is loop


try:
    from .yield_fixture_native import yield_fixture
except SyntaxError:
    from .yield_fixture import yield_fixture  # noqa


async def test_yield_fixture(yield_fixture):  # noqa
    assert yield_fixture is True


def test_localhost(localhost):
    assert socket.gethostbyname("localhost") == localhost



import asyncio
import socket
from functools import partial

from ..utils import OptionsType, awaitable, bind_socket
from .base import SimpleServer


class TCPServer(SimpleServer):
    PROTO_NAME = "tcp"

    def __init__(
        self, address: str = None, port: int = None,
        options: OptionsType = (), sock=None, **kwargs
    ):
        if not sock:
            if not (address and port):
                raise RuntimeError(
                    "You should pass socket instance or "
                    '"address" and "port" couple',
                )

            self.make_socket = partial(
                bind_socket,
                address=address,
                port=port,
                options=options,
            )
        elif not isinstance(sock, socket.socket):
            raise ValueError("sock must be socket instance")
        else:
            self.make_socket = lambda: sock

        self.socket = None

        super().__init__(**kwargs)

    async def handle_client(
        self, reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter
    ):
        raise NotImplementedError

    def make_client_handler(
        self, reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ):
        return self.create_task(awaitable(self.handle_client)(reader, writer))

    async def start(self):
        self.socket = self.make_socket()
        self.server = await asyncio.start_server(
            self.make_client_handler,
            sock=self.socket,
        )

    async def stop(self, exc: Exception = None):
        await super().stop(exc)
        await self.server.wait_closed()



import asyncio
import hashlib
import time

import pytest
from async_generator import async_generator, yield_

import aiomisc


class HashServer(aiomisc.service.TCPServer):
    async def handle_client(
        self, reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter
    ):
        hasher = hashlib.md5()

        chunk = await reader.read(65534)
        while chunk:
            hasher.update(chunk)
            writer.write(hasher.digest())
            chunk = await reader.read(65534)

        writer.close()
        await writer.wait_closed()


@pytest.fixture()
def server_port(aiomisc_unused_port_factory) -> int:
    return aiomisc_unused_port_factory()


@pytest.fixture()
def service(
    loop: asyncio.AbstractEventLoop,
    server_port, localhost,
) -> HashServer:
    return HashServer(port=server_port, address=localhost)


@pytest.fixture()
def services(service: HashServer):
    return [service]


@pytest.fixture()
@async_generator
async def proxy(tcp_proxy, localhost, server_port):
    async with tcp_proxy(localhost, server_port, buffered=True) as proxy:
        await yield_(proxy)


async def test_proxy_client(proxy):
    reader, writer = await proxy.create_client()
    payload = b"Hello world"

    writer.write(payload)
    hash = await asyncio.wait_for(reader.readexactly(16), timeout=1)

    assert hash == hashlib.md5(payload).digest()


async def test_proxy_client_close(proxy):
    reader, writer = await proxy.create_client()
    payload = b"Hello world"

    writer.write(payload)
    hash = await asyncio.wait_for(reader.readexactly(16), timeout=1)

    assert hash == hashlib.md5(payload).digest()

    assert not reader.at_eof()
    await proxy.disconnect_all()

    assert await asyncio.wait_for(reader.read(), timeout=1) == b""
    assert reader.at_eof()


async def test_proxy_client_slow(proxy):
    delay = 0.1
    proxy.set_delay(delay, delay)

    reader, writer = await proxy.create_client()
    payload = b"Hello world"

    delta = -time.monotonic()
    writer.write(payload)
    hash = await asyncio.wait_for(reader.readexactly(16), timeout=2)
    delta += time.monotonic()

    assert delta >= delay, [c.delay for c in proxy.clients]

    assert hash == hashlib.md5(payload).digest()


async def test_proxy_client_with_processor(proxy):
    processed_request = b"Never say hello"

    proxy.set_content_processors(
        lambda _: processed_request,
        lambda chunk: chunk[::-1],
    )

    reader, writer = await proxy.create_client()
    payload = b"Hello world"

    writer.write(payload)
    hash = await asyncio.wait_for(reader.read(16), timeout=2)

    assert hash == hashlib.md5(processed_request).digest()[::-1]



import logging

import pkg_resources

from . import io, log
from .backoff import asyncbackoff
from .context import Context, get_context
from .entrypoint import entrypoint
from .iterator_wrapper import IteratorWrapper
from .periodic import PeriodicCallback
from .pool import PoolBase
from .process_pool import ProcessPoolExecutor
from .service import Service
from .signal import Signal, receiver
from .thread_pool import (
    IteratorWrapperSeparate, ThreadPoolExecutor, context_partial,
    sync_wait_coroutine, threaded, threaded_iterable,
    threaded_iterable_separate, threaded_separate,
)
from .timeout import timeout
from .utils import (
    SelectResult, awaitable, bind_socket, cancel_tasks, chunk_list,
    new_event_loop, select, shield,
)


plugins = {
    entry_point.name: entry_point.load()
    for entry_point
    in pkg_resources.iter_entry_points("aiomisc.plugins")
}


def setup_plugins():
    logger = logging.getLogger(__name__)
    for name, plugin in plugins.items():
        try:
            logger.debug("Trying to load %r %r", name, plugin)
            plugin.setup()
        except:  # noqa
            logger.exception("Error on %s aiomisc plugin setup", name)
            raise


setup_plugins()


__all__ = (
    "asyncbackoff",
    "awaitable",
    "bind_socket",
    "cancel_tasks",
    "chunk_list",
    "Context",
    "context_partial",
    "entrypoint",
    "get_context",
    "io",
    "IteratorWrapper",
    "IteratorWrapperSeparate",
    "log",
    "new_event_loop",
    "PeriodicCallback",
    "plugins",
    "PoolBase",
    "ProcessPoolExecutor",
    "receiver",
    "select",
    "SelectResult",
    "Service",
    "shield",
    "Signal",
    "sync_wait_coroutine",
    "threaded",
    "threaded_iterable",
    "threaded_iterable_separate",
    "threaded_separate",
    "ThreadPoolExecutor",
    "timeout",
)



import asyncio
import logging
import typing

from .context import Context, get_context
from .log import LogFormat, basic_config
from .service import Service
from .signal import Signal
from .utils import create_default_event_loop, event_loop_policy


class Entrypoint:

    PRE_START = Signal()
    POST_STOP = Signal()

    async def _start(self):
        if self.log_config:
            basic_config(
                level=self.log_level,
                log_format=self.log_format,
                buffered=True,
                loop=self.loop,
                buffer_size=self.log_buffer_size,
                flush_interval=self.log_flush_interval,
            )

        for signal in (self.pre_start, self.post_stop):
            signal.freeze()

        await self.pre_start.call(entrypoint=self, services=self.services)

        await asyncio.gather(
            *[self._start_service(svc) for svc in self.services],
        )

    def __init__(
        self, *services, loop: asyncio.AbstractEventLoop = None,
        pool_size: int = None,
        log_level: typing.Union[int, str] = logging.INFO,
        log_format: typing.Union[str, LogFormat] = "color",
        log_buffer_size: int = 1024,
        log_flush_interval: float = 0.2,
        log_config: bool = True,
        policy=event_loop_policy,
        debug: bool = False
    ):

        """

        :param debug: set debug to event loop
        :param loop: loop
        :param services: Service instances which will be starting.
        :param pool_size: thread pool size
        :param log_level: Logging level which will be configured
        :param log_format: Logging format which will be configures
        :param log_buffer_size: Buffer size for logging
        :param log_flush_interval: interval in seconds for flushing logs
        :param log_config: if False do not configure logging
        """

        self._debug = debug
        self._loop = loop
        self._loop_owner = False
        self._thread_pool = None
        self.ctx = None
        self.log_buffer_size = log_buffer_size
        self.log_config = log_config
        self.log_flush_interval = log_flush_interval
        self.log_format = log_format
        self.log_level = log_level
        self.policy = policy
        self.pool_size = pool_size
        self.services = services
        self.shutting_down = False
        self.pre_start = self.PRE_START.copy()
        self.post_stop = self.POST_STOP.copy()

        self._closing = None

    async def closing(self):
        # Lazy initialization because event loop might be not exists
        if self._closing is None:
            self._closing = asyncio.Event()

        await self._closing.wait()

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if self._loop is None:
            self._loop, self._thread_pool = create_default_event_loop(
                pool_size=self.pool_size,
                policy=self.policy,
                debug=self._debug,
            )
            self._loop_owner = True

        return self._loop

    def __del__(self):
        if self._loop and self._loop.is_closed():
            return

        if self._loop_owner:
            self._loop.close()

    def __enter__(self) -> asyncio.AbstractEventLoop:
        self.loop.run_until_complete(self.__aenter__())
        return self.loop

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.loop.is_closed():
            return

        self.loop.run_until_complete(
            self.__aexit__(exc_type, exc_val, exc_tb),
        )

        if self._loop_owner:
            self._loop.close()

    async def __aenter__(self) -> "Entrypoint":
        if self._loop is None:
            # When __aenter__ called without __enter__
            self._loop = asyncio.get_event_loop()

        if self.log_config:
            basic_config(
                level=self.log_level,
                log_format=self.log_format,
                loop=self.loop,
                buffered=False,
            )

        self.ctx = Context(loop=self.loop)
        await self._start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        try:
            if self.loop.is_closed():
                return

            await self.graceful_shutdown(exc_val)
            self.shutting_down = True
        finally:
            self.ctx.close()

            if self._thread_pool:
                self._thread_pool.shutdown()

    async def _start_service(self, svc: Service):
        svc.set_loop(self.loop)

        start_task, ev_task = map(
            asyncio.ensure_future, (svc.start(), svc.start_event.wait()),
        )

        await asyncio.wait(
            (start_task, ev_task),
            return_when=asyncio.FIRST_COMPLETED,
        )

        self.loop.call_soon(svc.start_event.set)
        await ev_task

        if start_task.done():
            return await start_task

    async def graceful_shutdown(self, exception):
        if self._closing:
            self._closing.set()

        tasks = [
            asyncio.shield(svc.stop(exception)) for svc in self.services
        ]

        if tasks:
            await asyncio.gather(*tasks, return_exceptions=True)

        await self.post_stop.call(entrypoint=self)
        await self.loop.shutdown_asyncgens()


entrypoint = Entrypoint


__all__ = ("entrypoint", "Entrypoint", "get_context")



import asyncio
import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from random import random

try:
    from typing import AsyncContextManager
except ImportError:
    # Failed on Python 3.5.2 reproducible on ubuntu 16.04 (xenial)
    class AsyncContextManager(ABC):

        @abstractmethod
        async def __aenter__(self):
            raise NotImplementedError

        @abstractmethod
        async def __aexit__(self, exc_type, exc_val, exc_tb):
            raise NotImplementedError


from .utils import cancel_tasks


log = logging.getLogger(__name__)


class ContextManager(AsyncContextManager):
    __slots__ = "__aenter", "__aexit", "__instance"

    sentinel = object()

    def __init__(self, aenter, aexit):
        self.__aenter = aenter
        self.__aexit = aexit
        self.__instance = self.sentinel

    async def __aenter__(self):
        if self.__instance is not self.sentinel:
            raise RuntimeError("Reuse of context manager is not acceptable")

        self.__instance = await self.__aenter()
        return self.__instance

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.__aexit(self.__instance)


class PoolBase(ABC):
    __slots__ = (
        "_create_lock",
        "_instances",
        "_loop",
        "_recycle",
        "_recycle_bin",
        "_recycle_times",
        "_semaphore",
        "_tasks",
        "_len",
        "_used",
    )

    def __init__(self, maxsize=None, recycle=None):
        assert (
            recycle is None or recycle > 0
        ), "recycle should be positive number or None"

        self._loop = asyncio.get_event_loop()

        self._instances = asyncio.Queue()
        self._recycle_bin = asyncio.Queue()

        self._semaphore = asyncio.Semaphore(maxsize)
        self._len = 0

        self._recycle = recycle

        self._tasks = set()
        self._used = set()

        self._create_lock = asyncio.Lock()
        self._recycle_times = defaultdict(self._loop.time)

        self.__create_task(self.__recycler())

    def __create_task(self, coro):
        task = self._loop.create_task(coro)
        self._tasks.add(task)
        task.add_done_callback(self._tasks.remove)
        return task

    async def __recycler(self):
        while True:
            instance = await self._recycle_bin.get()

            try:
                await self._destroy_instance(instance)
            except Exception:
                log.exception("Error when recycle instance %r", instance)
            finally:
                self._recycle_bin.task_done()

    @abstractmethod
    async def _create_instance(self):
        pass

    @abstractmethod
    async def _destroy_instance(self, instance):
        pass

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    @abstractmethod
    async def _check_instance(self, instance):
        return True

    def __len__(self):
        return self._len

    def __recycle_instance(self, instance):
        self._len -= 1
        self._semaphore.release()
        if instance in self._recycle_times:
            self._recycle_times.pop(instance)

        if instance in self._used:
            self._used.remove(instance)

        self._recycle_bin.put_nowait(instance)

    async def __create_new_instance(self):
        await self._semaphore.acquire()

        instance = await self._create_instance()
        self._len += 1

        if self._recycle:
            deadline = self._recycle * (1 + random())
            self._recycle_times[instance] += deadline

        await self._instances.put(instance)

    async def __acquire(self):
        if not self._semaphore.locked():
            await self.__create_new_instance()

        instance = await self._instances.get()

        try:
            result = await self._check_instance(instance)
        except Exception:
            log.exception("Check instance %r failed", instance)
            self.__recycle_instance(instance)
        else:
            if not result:
                self.__recycle_instance(instance)
                return await self.__acquire()

        self._used.add(instance)
        return instance

    async def __release(self, instance):
        self._used.remove(instance)

        if self._recycle and self._recycle_times[instance] < self._loop.time():
            self.__recycle_instance(instance)
            return

        self._instances.put_nowait(instance)

    def acquire(self):
        return ContextManager(self.__acquire, self.__release)

    async def close(self, timeout=None):
        instances = list(self._used)
        self._used.clear()

        while self._instances.qsize():
            try:
                instances.append(self._instances.get_nowait())
            except asyncio.QueueEmpty:
                break

        async def log_exception(coro):
            try:
                await coro
            except Exception:
                log.exception("Exception when task execution")

        await asyncio.wait_for(
            asyncio.gather(
                *[
                    self.__create_task(
                        log_exception(self._destroy_instance(instance)),
                    )
                    for instance in instances
                ],
                return_exceptions=True
            ),
            timeout=timeout,
        )

        await cancel_tasks(self._tasks)



import os

from setuptools import setup, find_packages


from importlib.machinery import SourceFileLoader


module_name = 'aiomisc'

try:
    version = SourceFileLoader(
        module_name,
        os.path.join(module_name, 'version.py')
    ).load_module()

    version_info = version.version_info
except FileNotFoundError:
    version_info = (0, 0, 0)


__version__ = '{}.{}.{}'.format(*version_info)


def load_requirements(fname):
    """ load requirements from a pip requirements file """
    with open(fname) as f:
        line_iter = (line.strip() for line in f.readlines())
        return [line for line in line_iter if line and line[0] != '#']


setup(
    name=module_name,
    version=__version__,
    author='Dmitry Orlov',
    author_email='me@mosquito.su',
    license='MIT',
    description='aiomisc - miscellaneous utils for asyncio',
    long_description=open("README.rst").read(),
    platforms="all",
    classifiers=[
        "Framework :: Pytest",
        'Intended Audience :: Developers',
        'Natural Language :: Russian',
        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    packages=find_packages(exclude=['tests']),
    install_requires=load_requirements('requirements.txt'),
    extras_require={
        'aiohttp': ['aiohttp'],
        'carbon': ['aiocarbon~=0.15'],
        'contextvars': ['contextvars~=2.4'],
        'develop': load_requirements('requirements.dev.txt'),
        'raven': ['raven-aiohttp'],
        'uvloop': ['uvloop>=0.14,<1'],
        'cron': ['croniter~=0.3.34'],
        ':python_version < "3.7"': 'async-generator',
    },
    entry_points={
        "pytest11": ["aiomisc = aiomisc_pytest.pytest_plugin"]
    },
    url='https://github.com/mosquito/aiomisc'
)
